package ru.dbotthepony.kommons.io

import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.*
import kotlin.reflect.KClass

/**
 * Represents value which can be encoded onto or decoded from stream.
 *
 * Also provides [copy] and [compare] methods
 */
interface StreamCodec<V> {
	fun read(stream: DataInputStream): V
	fun write(stream: DataOutputStream, value: V)

	/**
	 * defensive copy
	 */
	fun copy(value: V): V

	/**
	 * Optional equality check override. Utilized to determine whenever e.g. network value is different from new value
	 *
	 * By default uses [Any.equals]
	 */
	fun compare(a: V, b: V): Boolean {
		return a == b
	}

	class Impl<V>(
		private val reader: (stream: DataInputStream) -> V,
		private val writer: (stream: DataOutputStream, value: V) -> Unit,
		private val copier: ((value: V) -> V) = { it },
		private val comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
	) : StreamCodec<V> {
		constructor(
			reader: (stream: DataInputStream) -> V,
			payloadSize: Long,
			writer: (stream: DataOutputStream, value: V) -> Unit,
			copier: ((value: V) -> V) = { it },
			comparator: ((a: V, b: V) -> Boolean) = { a, b -> a == b }
		) : this({ stream -> reader.invoke(stream) }, writer, copier, comparator)

		val nullable = object : StreamCodec<V?> {
			override fun read(stream: DataInputStream): V? {
				return if (stream.read() == 0) null else reader.invoke(stream)
			}

			override fun write(stream: DataOutputStream, value: V?) {
				if (value === null) stream.write(0) else {
					stream.write(1)
					writer.invoke(stream, value)
				}
			}

			override fun copy(value: V?): V? {
				return if (value === null) null else copier.invoke(value)
			}

			override fun compare(a: V?, b: V?): Boolean {
				if (a === null && b === null) return true
				if (a === null || b === null) return false
				return comparator.invoke(a, b)
			}
		}

		override fun read(stream: DataInputStream): V {
			return reader.invoke(stream)
		}

		override fun write(stream: DataOutputStream, value: V) {
			writer.invoke(stream, value)
		}

		override fun copy(value: V): V {
			return copier.invoke(value)
		}

		override fun compare(a: V, b: V): Boolean {
			return comparator.invoke(a, b)
		}
	}

	class Collection<E, C : MutableCollection<E>>(val elementCodec: StreamCodec<E>, val collectionFactory: (Int) -> C) : StreamCodec<C> {
		override fun read(stream: DataInputStream): C {
			val size = stream.readVarInt()

			if (size <= 0) {
				return collectionFactory.invoke(0)
			}

			val collection = collectionFactory.invoke(size)

			for (i in 0 until size) {
				collection.add(elementCodec.read(stream))
			}

			return collection
		}

		override fun write(stream: DataOutputStream, value: C) {
			stream.writeVarInt(value.size)
			value.forEach { elementCodec.write(stream, it) }
		}

		override fun copy(value: C): C {
			val new = collectionFactory.invoke(value.size)
			value.forEach { new.add(elementCodec.copy(it)) }
			return new
		}
	}

	class Enum<V : kotlin.Enum<V>>(clazz: Class<out V>) : StreamCodec<V> {
		val clazz = searchClass(clazz)
		val values: List<V> = listOf(*this.clazz.enumConstants!!)
		val valuesMap = values.associateBy { it.name }

		override fun read(stream: DataInputStream): V {
			val id = stream.readVarInt()
			return values.getOrNull(id) ?: throw NoSuchElementException("No such enum with index $id")
		}

		override fun write(stream: DataOutputStream, value: V) {
			stream.writeVarInt(value.ordinal)
		}

		override fun copy(value: V): V {
			return value
		}

		override fun compare(a: V, b: V): Boolean {
			return a === b
		}

		companion object {
			/**
			 * FIXME: enums with abstract methods which get compiled to subclasses, whose DO NOT expose "parent's" enum constants array
			 *
			 * is there an already existing solution?
			 */
			fun <V : kotlin.Enum<V>> searchClass(clazz: Class<out V>): Class<out V> {
				var search: Class<*> = clazz

				while (search.enumConstants == null && search.superclass != null) {
					search = search.superclass
				}

				if (search.enumConstants == null) {
					throw ClassCastException("$clazz does not represent an enum or enum subclass")
				}

				return search as Class<out V>
			}
		}
	}
}

val NullValueCodec = StreamCodec.Impl({ _ -> null }, { _, _ -> })
val BooleanValueCodec = StreamCodec.Impl(DataInputStream::readBoolean, 1L, DataOutputStream::writeBoolean)
val ByteValueCodec = StreamCodec.Impl(DataInputStream::readByte, 1L, { s, v -> s.writeByte(v.toInt()) })
val ShortValueCodec = StreamCodec.Impl(DataInputStream::readShort, 2L, { s, v -> s.writeShort(v.toInt()) })
val CharValueCodec = StreamCodec.Impl(DataInputStream::readChar, 2L, { s, v -> s.writeShort(v.code) })
val IntValueCodec = StreamCodec.Impl(DataInputStream::readInt, 4L, DataOutputStream::writeInt)
val LongValueCodec = StreamCodec.Impl(DataInputStream::readLong, 8L, DataOutputStream::writeLong)
val FloatValueCodec = StreamCodec.Impl(DataInputStream::readFloat, 4L, DataOutputStream::writeFloat)
val DoubleValueCodec = StreamCodec.Impl(DataInputStream::readDouble, 8L, DataOutputStream::writeDouble)
val BigDecimalValueCodec = StreamCodec.Impl(DataInputStream::readBigDecimal, DataOutputStream::writeBigDecimal)
val UUIDValueCodec = StreamCodec.Impl({ s -> UUID(s.readLong(), s.readLong()) }, { s, v -> s.writeLong(v.mostSignificantBits); s.writeLong(v.leastSignificantBits) })
val VarIntValueCodec = StreamCodec.Impl(DataInputStream::readSignedVarInt, DataOutputStream::writeSignedVarInt)
val VarLongValueCodec = StreamCodec.Impl(DataInputStream::readSignedVarLong, DataOutputStream::writeSignedVarLong)
val BinaryStringCodec = StreamCodec.Impl(DataInputStream::readBinaryString, DataOutputStream::writeBinaryString)

fun <E : Enum<E>> Class<E>.codec() = StreamCodec.Enum(this)
fun <E : Enum<E>> KClass<E>.codec() = StreamCodec.Enum(this.java)
