@file:Suppress("DeprecatedCallableAddReplaceWith")

package ru.dbotthepony.kommons.io

import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceArraySet
import ru.dbotthepony.kommons.collect.ProxiedMap
import ru.dbotthepony.kommons.collect.forValidRefs
import ru.dbotthepony.kommons.util.Observer
import ru.dbotthepony.kommons.util.ListenableDelegate
import ru.dbotthepony.kommons.util.DelegateGetter
import ru.dbotthepony.kommons.util.DelegateSetter
import ru.dbotthepony.kommons.util.Listenable
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.lang.ref.WeakReference
import java.util.*
import java.util.function.Consumer
import java.util.function.Supplier

/**
 * Universal, one-to-many value synchronizer, allowing to synchronize values from server to client
 * anywhere, where input/output streams are supported
 */
@Suppress("unused")
class SynchedDelegates(private val callback: Runnable, private val alwaysCallCallback: Boolean) {
	constructor() : this(Runnable {}, false)
	constructor(callback: Runnable) : this(callback, false)

	private var freeSlots = 0
	// почему не удалять поля напрямую?
	// чтоб не возникло проблем в состоянии гонки
	// формируем пакет -> удаляем поле по обе стороны -> клиент принимает пакет -> клиент считывает неверные данные
	// конечно, всё равно всё сломается если было удалено поле, которое находится в пакете
	// но если поля нет в пакете, то всё окей
	private val fields = ArrayList<AbstractValue?>(0)
	private val observers = ArrayList<Observer>(0)
	private val endpoints = ArrayList<WeakReference<Endpoint>>(1)
	private var lastEndpointCleanup = System.nanoTime()
	private var nextFieldID = 0
	val hasObservers: Boolean get() = observers.isNotEmpty()
	var isEmpty: Boolean = true
		private set
	val isNotEmpty: Boolean get() = !isEmpty
	var isDirty: Boolean = false
		private set(value) {
			if (value != field) {
				field = value

				if (value && !alwaysCallCallback) {
					callback.run()
				}
			}

			if (alwaysCallCallback && value) {
				callback.run()
			}
		}

	private val boundEndpoints = Collections.synchronizedMap(WeakHashMap<Any, Endpoint>())

	fun computeEndpointFor(obj: Any): Endpoint {
		return boundEndpoints.computeIfAbsent(obj) { Endpoint() }
	}

	fun removeEndpointFor(obj: Any): Endpoint? {
		return boundEndpoints.remove(obj)
	}

	fun endpointFor(obj: Any): Endpoint? {
		return boundEndpoints[obj]
	}

	fun markClean() {
		isDirty = false
	}

	private fun notifyEndpoints(dirtyField: AbstractValue) {
		isDirty = true

		forEachEndpoint {
			it.addDirty(dirtyField)
		}
	}

	private inline fun forEachEndpoint(execute: (Endpoint) -> Unit) {
		synchronized(endpoints) {
			endpoints.forValidRefs { execute.invoke(it) }
		}
	}

	val defaultEndpoint = Endpoint()

	inner class Endpoint {
		init {
			synchronized(endpoints) {
				endpoints.add(WeakReference(this))

				if (System.nanoTime() - lastEndpointCleanup >= 60) {
					lastEndpointCleanup = System.nanoTime()

					val iterator = endpoints.listIterator()

					for (value in iterator) {
						if (value.get() == null) {
							iterator.remove()
						}
					}
				}
			}
		}

		private val dirty = ReferenceArraySet<AbstractValue>(4)

		// use LinkedList because it is ensured memory is freed on LinkedList#clear
		private val mapBacklogs = Reference2ObjectOpenHashMap<Map<*, *>, LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()
		private val setBacklogs = Reference2ObjectOpenHashMap<Set<*>, LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

		var isDisabled: Boolean = false
			private set

		fun disable() {
			if (isDisabled) return
			isDisabled = true
			mapBacklogs.clear()
			dirty.clear()

			synchronized(endpoints) {
				val iterator = endpoints.listIterator()

				for (value in iterator) {
					if (value.get() === this || value.get() == null) {
						iterator.remove()
					}
				}
			}
		}

		init {
			markDirty()
		}

		fun markDirty() {
			if (isDisabled) return

			for (field in fields) {
				field?.markDirty(this)
			}
		}

		internal fun addDirty(field: AbstractValue) {
			if (isDisabled) return
			dirty.add(field)
		}

		internal fun removeDirty(field: AbstractValue) {
			dirty.remove(field)
		}

		internal fun <K, V> getMapBacklog(map: Map<K, V>): LinkedList<Pair<Any?, (DataOutputStream) -> Unit>> {
			if (isDisabled) return LinkedList()

			return mapBacklogs.computeIfAbsent(map, Reference2ObjectFunction {
				LinkedList()
			})
		}

		internal fun <K, V> removeMapBacklog(map: Map<K, V>) {
			mapBacklogs.remove(map)
		}

		internal fun <V> getSetBacklog(set: Set<V>): LinkedList<Pair<Any?, (DataOutputStream) -> Unit>> {
			if (isDisabled) return LinkedList()

			return setBacklogs.computeIfAbsent(set, Reference2ObjectFunction {
				LinkedList()
			})
		}

		internal fun <V> removeSetBacklog(set: Set<V>) {
			setBacklogs.remove(set)
		}

		fun collectData(): FastByteArrayOutputStream? {
			if (isDisabled || dirty.isEmpty())
				return null

			val stream = FastByteArrayOutputStream()
			val dataStream = DataOutputStream(stream)

			for (field in dirty) {
				stream.writeVarInt(field.id)
				field.write(dataStream, this)
			}

			dirty.clear()
			stream.write(0)

			return stream
		}
	}

	abstract inner class AbstractValue : Observer {
		val id: Int

		init {
			if (freeSlots > 0) {
				var found = -1

				for (i in fields.indices) {
					if (fields[i] == null) {
						fields[i] = this
						found = i + 1
						freeSlots--
						break
					}
				}

				if (found == -1) {
					throw RuntimeException("freeSlots = $freeSlots but no null entries in field list!")
				} else {
					id = found
				}
			} else {
				fields.add(this)
				id = fields.size
				isEmpty = false
			}
		}

		var isRemoved = false
			private set

		protected var isDirty = false
		protected var isRemote = false

		open fun remove() {
			if (isRemoved)
				return

			isRemoved = true
			freeSlots++
			fields[id - 1] = null
			observers.remove(this)
			isEmpty = fields.all { it == null }

			while (fields[fields.size - 1] == null) {
				fields.removeAt(fields.size - 1)
				freeSlots--
			}

			forEachEndpoint {
				it.removeDirty(this)
			}
		}

		open fun markDirty(endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }
			endpoint.addDirty(this)
		}

		open fun markDirty() {
			check(!isRemoved) { "Field was removed" }
			notifyEndpoints(this)
			isDirty = true
		}

		abstract fun write(stream: DataOutputStream, endpoint: Endpoint)
		abstract fun read(stream: DataInputStream)
	}

	inner class RegularValue<T>(
		value: T,
		private val codec: StreamCodec<T>,
		getter: DelegateGetter<T> = DelegateGetter.passthrough(),
		setter: DelegateSetter<T> = DelegateSetter.passthrough()
	) : AbstractValue(), ListenableDelegate<T> by ListenableDelegate.maskSmart(value, getter, setter) {
		init {
			addListener(Runnable {
				if (!isDirty && !isRemote) {
					notifyEndpoints(this)
					isDirty = true
				}
			})
		}

		override fun observe(): Boolean {
			return false
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			codec.write(stream, get())
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			isRemote = true
			accept(codec.read(stream))
			isDirty = false
		}
	}

	inner class ObservedValue<T>(
		value: T,
		private val codec: StreamCodec<T>,
		getter: DelegateGetter<T> = DelegateGetter.passthrough(),
		setter: DelegateSetter<T> = DelegateSetter.passthrough()
	) : AbstractValue(), ListenableDelegate<T> by ListenableDelegate.maskSmart(value, getter, setter) {
		private var observed = codec.copy(value)

		init {
			observers.add(this)

			addListener(Runnable {
				if (!isDirty && !isRemote) {
					notifyEndpoints(this)
					isDirty = true
				}
			})
		}

		override fun observe(): Boolean {
			if (!isRemote) {
				val get = get()

				if (!codec.compare(get, observed)) {
					observed = codec.copy(get)
					isDirty = true
				}
			}

			return isDirty
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			codec.write(stream, get())
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			isRemote = true
			accept(codec.read(stream))
			isDirty = false
		}
	}

	inner class ComputedValue<T>(getter: Supplier<T>, private val codec: StreamCodec<T>) : AbstractValue(), ListenableDelegate<T> {
		private val shadow = ListenableDelegate.SmartShadow(getter, codec::compare, codec::copy)

		override fun get(): T {
			return shadow.get()
		}

		override fun accept(t: T) {
			shadow.accept(t)
		}

		override fun addListener(listener: Consumer<T>): Listenable.L {
			return shadow.addListener(listener)
		}

		override fun observe(): Boolean {
			return shadow.observe()
		}

		init {
			shadow.addListener(Runnable {
				if (!isDirty && !isRemote) {
					notifyEndpoints(this)
					isDirty = true
				}
			})
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			codec.write(stream, shadow.get())
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			isRemote = true
			shadow.accept(codec.read(stream))
			isDirty = false
		}
	}

	@JvmName("vbyte") fun byte(value: Byte = 0, getter: DelegateGetter<Byte> = DelegateGetter.passthrough(), setter: DelegateSetter<Byte> = DelegateSetter.passthrough()): ListenableDelegate<Byte> = RegularValue(value, ByteValueCodec, getter, setter)
	@JvmName("vshort") fun short(value: Short = 0, getter: DelegateGetter<Short> = DelegateGetter.passthrough(), setter: DelegateSetter<Short> = DelegateSetter.passthrough()): ListenableDelegate<Short> = RegularValue(value, ShortValueCodec, getter, setter)
	@JvmName("vchar") fun char(value: Char = 0.toChar(), getter: DelegateGetter<Char> = DelegateGetter.passthrough(), setter: DelegateSetter<Char> = DelegateSetter.passthrough()): ListenableDelegate<Char> = RegularValue(value, CharValueCodec, getter, setter)
	@JvmName("vint") fun int(value: Int = 0, getter: DelegateGetter<Int> = DelegateGetter.passthrough(), setter: DelegateSetter<Int> = DelegateSetter.passthrough()): ListenableDelegate<Int> = RegularValue(value, IntValueCodec, getter, setter)
	@JvmName("vlong") fun long(value: Long = 0L, getter: DelegateGetter<Long> = DelegateGetter.passthrough(), setter: DelegateSetter<Long> = DelegateSetter.passthrough()): ListenableDelegate<Long> = RegularValue(value, LongValueCodec, getter, setter)
	@JvmName("vfloat") fun float(value: Float = 0f, getter: DelegateGetter<Float> = DelegateGetter.passthrough(), setter: DelegateSetter<Float> = DelegateSetter.passthrough()): ListenableDelegate<Float> = RegularValue(value, FloatValueCodec, getter, setter)
	@JvmName("vdouble") fun double(value: Double = 0.0, getter: DelegateGetter<Double> = DelegateGetter.passthrough(), setter: DelegateSetter<Double> = DelegateSetter.passthrough()): ListenableDelegate<Double> = RegularValue(value, DoubleValueCodec, getter, setter)
	@JvmName("vboolean") fun boolean(value: Boolean = false, getter: DelegateGetter<Boolean> = DelegateGetter.passthrough(), setter: DelegateSetter<Boolean> = DelegateSetter.passthrough()): ListenableDelegate<Boolean> = RegularValue(value, BooleanValueCodec, getter, setter)

	fun string(value: String, getter: DelegateGetter<String> = DelegateGetter.passthrough(), setter: DelegateSetter<String> = DelegateSetter.passthrough()): ListenableDelegate<String> = RegularValue(value, BinaryStringCodec, getter, setter)
	fun uuid(value: UUID, getter: DelegateGetter<UUID> = DelegateGetter.passthrough(), setter: DelegateSetter<UUID> = DelegateSetter.passthrough()): ListenableDelegate<UUID> = RegularValue(value, UUIDValueCodec, getter, setter)

	inner class Set<E>(
		private val codec: StreamCodec<E>,
		private val backingSet: MutableSet<E>,
		private val callback: ((changes: Collection<SetChangeset<E>>) -> Unit)? = null,
	) : AbstractValue() {
		private fun pushBacklog(element: E, action: (DataOutputStream) -> Unit) {
			check(!isRemote) { "Field marked as remote" }

			val pair = element to action

			forEachEndpoint {
				val list = it.getSetBacklog(this)
				val iterator = list.listIterator()

				for (value in iterator) {
					if (value.first == element) {
						iterator.remove()
					}
				}

				list.addLast(pair)
			}
		}

		override fun observe(): Boolean {
			return false
		}

		override fun remove() {
			if (!isRemoved) {
				forEachEndpoint { it.removeSetBacklog(this) }
			}

			super.remove()
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }

			if (isRemote || endpoints.isEmpty())
				return

			isDirty = true

			val endpoints = LinkedList<MutableList<Pair<Any?, (DataOutputStream) -> Unit>>>()

			forEachEndpoint {
				endpoints.add(it.getSetBacklog(this))
				it.addDirty(this)
			}

			endpoints.forEach {
				it.clear()
				it.add(null to ClearBacklogEntry)
			}

			for (value in backingSet) {
				val valueCopy = codec.copy(value)
				val pair: Pair<Any?, (DataOutputStream) -> Unit> = valueCopy to { it.write(ChangesetAction.ADD.ordinal + 1); codec.write(it, valueCopy) }

				endpoints.forEach {
					it.add(pair)
				}
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			super.markDirty(endpoint)

			endpoint.getSetBacklog(this).let {
				it.clear()
				it.add(null to ClearBacklogEntry)

				for (value in backingSet) {
					val valueCopy = codec.copy(value)
					it.add(valueCopy to { it.write(ChangesetAction.ADD.ordinal + 1); codec.write(it, valueCopy) })
				}
			}
		}

		val value: MutableSet<E> = object : MutableSet<E> {
			override fun add(element: E): Boolean {
				if (backingSet.add(element)) {
					if (!isRemote) {
						markDirty()

						val copy = codec.copy(element)

						pushBacklog(element) {
							it.write(ChangesetAction.ADD.ordinal + 1)
							codec.write(it, copy)
						}
					}

					return true
				}

				return false
			}

			override fun addAll(elements: Collection<E>): Boolean {
				var any = false
				elements.forEach { any = add(it) || any }
				return any
			}

			override fun clear() {
				if (backingSet.isNotEmpty()) {
					backingSet.clear()

					if (!isRemote) {
						markDirty()

						forEachEndpoint {
							it.getSetBacklog(this@Set).let {
								it.clear()
								it.add(null to ClearBacklogEntry)
							}
						}
					}
				}
			}

			override fun iterator(): MutableIterator<E> {
				return object : MutableIterator<E> {
					private val parent = backingSet.iterator()
					private var lastElement: Any? = Mark

					override fun hasNext(): Boolean {
						return parent.hasNext()
					}

					override fun next(): E {
						return parent.next().also { lastElement = it }
					}

					override fun remove() {
						parent.remove()
						val lastElement = lastElement

						if (lastElement !== Mark) {
							this.lastElement = Mark

							if (!isRemote) {
								markDirty()

								@Suppress("unchecked_cast")
								pushBacklog(lastElement as E) {
									it.write(ChangesetAction.REMOVE.ordinal + 1)
									codec.write(it, lastElement as E)
								}
							}
						}
					}
				}
			}

			override fun remove(element: E): Boolean {
				if (backingSet.remove(element)) {
					if (!isRemote) {
						markDirty()

						val copy = codec.copy(element)

						pushBacklog(element) {
							it.write(ChangesetAction.REMOVE.ordinal + 1)
							codec.write(it, copy)
						}
					}

					return true
				}

				return false
			}

			override fun removeAll(elements: Collection<E>): Boolean {
				var any = false
				elements.forEach { any = remove(it) || any }
				return any
			}

			override fun retainAll(elements: Collection<E>): Boolean {
				var any = false

				val iterator = iterator()

				for (value in iterator) {
					if (value !in elements) {
						any = true
						iterator.remove()
					}
				}

				return any
			}

			override val size: Int
				get() = backingSet.size

			override fun contains(element: E): Boolean {
				return element in backingSet
			}

			override fun containsAll(elements: Collection<E>): Boolean {
				return backingSet.containsAll(elements)
			}

			override fun isEmpty(): Boolean {
				return backingSet.isEmpty()
			}
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			val list = endpoint.getSetBacklog(this)

			for (value in list) {
				value.second.invoke(stream)
			}

			stream.write(0)
			list.clear()
			isDirty = false
		}

		override fun read(stream: DataInputStream) {
			if (!isRemote) {
				isRemote = true
				forEachEndpoint { it.removeSetBacklog(this) }
			}

			isDirty = false

			var action = stream.read()
			val changeset = LinkedList<SetChangeset<E>>()

			while (action != 0) {
				if ((action - 1) !in ChangesetActionList.indices) {
					throw IllegalArgumentException("Unknown changeset action with index ${action - 1}")
				}

				when (ChangesetActionList[action - 1]) {
					ChangesetAction.CLEAR -> {
						changeset.add(SetChangeset(ChangesetAction.CLEAR, null))
						backingSet.clear()
					}

					ChangesetAction.ADD -> {
						val read = codec.read(stream)
						changeset.add(SetChangeset(ChangesetAction.ADD, read))
						backingSet.add(read)
					}

					ChangesetAction.REMOVE -> {
						val read = codec.read(stream)
						changeset.add(SetChangeset(ChangesetAction.REMOVE, read))
						backingSet.remove(read)
					}
				}

				action = stream.read()
			}

			callback?.invoke(changeset)
		}
	}

	inner class Map<K, V>(
		private val keyCodec: StreamCodec<K>,
		private val valueCodec: StreamCodec<V>,
		private val backingMap: MutableMap<K, V>,
		private val callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null,
	) : AbstractValue() {
		private var sentAllValues = false

		private fun pushBacklog(key: Any?, value: (DataOutputStream) -> Unit) {
			val pair = key to value

			forEachEndpoint {
				val list = it.getMapBacklog(this)
				val iterator = list.listIterator()

				for (e in iterator) {
					if (e.first == key) {
						iterator.remove()
					}
				}

				list.addLast(pair)
			}
		}

		override fun observe(): Boolean {
			return false
		}

		override fun markDirty() {
			check(!isRemoved) { "Field was removed" }

			if (isRemote || endpoints.isEmpty())
				return

			isDirty = true
			val backlogs = LinkedList<LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

			forEachEndpoint {
				it.addDirty(this)
				val value = it.getMapBacklog(this)
				backlogs.add(value)
				value.clear()
				value.add(null to ClearBacklogEntry)
			}

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				val action = key to { it: DataOutputStream ->
					it.write(ChangesetAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				for (backlog in backlogs) {
					backlog.add(action)
				}
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			check(!isRemoved) { "Field was removed" }

			if (isRemote)
				return

			val backlog = endpoint.getMapBacklog(this)

			backlog.clear()
			backlog.add(null to ClearBacklogEntry)

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				backlog.add(key to {
					it.write(ChangesetAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				})
			}
		}

		private fun lmarkDirty() {
			if (!isDirty) {
				notifyEndpoints(this@Map)
				isDirty = true
			}
		}

		val value: MutableMap<K, V> = object : ProxiedMap<K, V>(backingMap) {
			override fun onClear() {
				if (isRemote || endpoints.isEmpty())
					return

				forEachEndpoint { endpoint ->
					endpoint.getMapBacklog(this@Map).let {
						it.clear()
						it.add(null to ClearBacklogEntry)
					}
				}

				lmarkDirty()
				this@SynchedDelegates.isDirty = true
			}

			override fun onValueAdded(key: K, value: V) {
				if (isRemote || endpoints.isEmpty())
					return

				val keyCopy = keyCodec.copy(key)
				val valueCopy = valueCodec.copy(value)

				pushBacklog(key) {
					it.write(ChangesetAction.ADD.ordinal + 1)
					keyCodec.write(it, keyCopy)
					valueCodec.write(it, valueCopy)
				}

				lmarkDirty()
			}

			override fun onValueRemoved(key: K, value: V) {
				if (isRemote || endpoints.isEmpty())
					return

				val keyCopy = keyCodec.copy(key)

				pushBacklog(key) {
					it.write(ChangesetAction.REMOVE.ordinal + 1)
					keyCodec.write(it, keyCopy)
				}

				lmarkDirty()
			}
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			sentAllValues = false
			isDirty = false

			val iterator = endpoint.getMapBacklog(this).listIterator()

			for (entry in iterator) {
				entry.second.invoke(stream)
				iterator.remove()
			}

			stream.write(0)
		}

		override fun read(stream: DataInputStream) {
			if (!isRemote) {
				isRemote = true
				forEachEndpoint { it.removeMapBacklog(this) }
			}

			isDirty = false

			val changeset = LinkedList<MapChangeset<K, V>>()
			var readAction = stream.read() - 1

			while (readAction != -1) {
				if (readAction >= ChangesetActionList.size) {
					throw IndexOutOfBoundsException("Unknown map action with ID $readAction")
				}

				when (ChangesetActionList[readAction]) {
					ChangesetAction.CLEAR -> {
						backingMap.clear()
						changeset.add(ClearMapChangeset)
					}

					ChangesetAction.ADD -> {
						val key = keyCodec.read(stream)
						val value = valueCodec.read(stream)
						backingMap[key] = value
						changeset.add(MapChangeset(ChangesetAction.ADD, key, value))
					}

					ChangesetAction.REMOVE -> {
						val key = keyCodec.read(stream)
						backingMap.remove(key)
						changeset.add(MapChangeset(ChangesetAction.REMOVE, key, null))
					}
				}

				readAction = stream.read() - 1
			}

			if (changeset.size != 0) {
				callback?.invoke(changeset)
			}
		}
	}

	/**
	 * marks all fields dirty, invalidates mappings for each endpoint
	 */
	fun invalidate() {
		for (field in fields) {
			field?.markDirty()
		}
	}

	/**
	 * Observe changes of all fields with backing computation lambda
	 */
	fun observe(): Boolean {
		var changes = false

		if (observers.isNotEmpty()) {
			for (field in observers) {
				changes = field.observe() || changes
			}
		}

		return changes
	}

	/**
	 * [defaultEndpoint]#collectNetworkPayload
	 */
	fun collectData(): FastByteArrayOutputStream? {
		check(!defaultEndpoint.isDisabled) { "Default endpoint is not used" }
		observe()
		val values = defaultEndpoint.collectData()
		markClean()
		return values
	}

	fun read(stream: InputStream): Int {
		var fieldId = stream.readVarInt()
		var i = 0
		val dataStream = DataInputStream(stream)

		while (fieldId != 0) {
			val field = fields.getOrNull(fieldId - 1) ?: throw IllegalArgumentException("Unknown field ID ${fieldId - 1}")
			field.read(dataStream)
			fieldId = stream.readVarInt()
			i++
		}

		return i
	}

	private object Mark

	companion object {
		private val ClearBacklogEntry = { stream: DataOutputStream -> stream.write(ChangesetAction.CLEAR.ordinal + 1) }
		private val ChangesetActionList = ChangesetAction.values()
		private val ClearMapChangeset = MapChangeset(ChangesetAction.CLEAR, null, null)
	}
}
